#include <arpa/inet.h>
#include <assert.h>
#include <ctype.h>
#include <fcntl.h>
#include <limits.h>
#include <math.h>
#include <pwd.h>
#include <signal.h>
#include <string.h>
#include <stdarg.h>
#include <sys/dir.h>
#include <sys/ioctl.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <syslog.h>
#include <utime.h>
#include <time.h>   // Needed for radutmp.h
#include "radutmp.h"
#include "common.h"

#define	USAGE_TOTAL		0
#define	USAGE_DAY		1
#define	USAGE_WEEK		2
#define	USAGE_MONTH		3
#define	USAGE_YEAR		4

/* prefix for local installs into /usr/local for example */
#ifndef PREFIX
#define PREFIX ""
#endif

void                     update(word myHostAddr, time_t t, int effect, int loggingLevel, int timer);

int			 sendMail(UserRec *ur, char *login, char *cc, char *subject, char *sPath);

int			 warnExpire(time_t expire);
void			 writeUsageRecords(int curMinute);
void			 unlinkOldLockFiles();
void                     WarnExpire(UserRec *ur, char *login);

// *************************************************************************

int
main(int argc, char **argv)
{
  int                      i,
                           automatic = 1,     // do nightly maint.
                           runAsDaemon = 1,   // run as a daemon.
                           effect = 1,        // if 0, dont do anything.
			   timer = 1,	      // timer enable/disable.
                           loggingLevel = 1;  // more verbose logging.
  struct stat              _stat;
  struct tm               *tmP;
  struct in_addr           _in_addr;
  char                     hostname[256];
  time_t                   t,
                           configTime;
  word                     myHostAddr;
  char			  *_argv[16] = { "acua", NULL, NULL };
  char			   arg[16];

  readConfig();
  configTime = time(NULL);
  
  while ((i = getopt(argc, argv, "adelt")) != EOF) {
    switch (i) {
    case 'a':
      automatic = 0;
      break;
    case 'd':
      runAsDaemon = 0;
      break;
    case 'e':
      effect = 0;
      break;
    case 't':
      timer = 0;
      break;
    case 'l':
      loggingLevel = atoi(argv[optind++]);
      break;
    }
  }

  openlog("acua_radupdated", LOG_PID, LOG_DAEMON);
  
  gethostname(hostname, 256);
  myHostAddr = hostAddr(hostname);
  
  _in_addr.s_addr = myHostAddr;
  if (loggingLevel >= 1)
    syslog(LOG_INFO, "hostname = %s, IP = %s",
           hostname,
           inet_ntoa(_in_addr));

  // announce the modes acua is running in.
  syslog(LOG_INFO, "Current settings:");
  if (automatic) {
    syslog(LOG_INFO, "Maint: Yes");
  } else {
    syslog(LOG_INFO, "Maint: No");
  }
  if (runAsDaemon) {
    syslog(LOG_INFO, "Daemon: Yes");
  } else {
    syslog(LOG_INFO, "Daemon: No");
  }
  if (effect) {
    syslog(LOG_INFO, "Effect: Yes");
  } else {
    syslog(LOG_INFO, "Effect: No");
  }
  syslog(LOG_INFO, "Logging Level: %d", loggingLevel);

  if (runAsDaemon) daemonInit();

  for (;;) {
    // re-read the config file if it's been updated
    if (stat(configFilePath, &_stat) < 0)
      perrQuit("couldn't stat %s", configFilePath);
    if (_stat.st_mtime > configTime) {
      readConfig();
      if (configTime && loggingLevel >= 1)
        syslog(LOG_INFO, "re-read %s", configFilePath);
      configTime = _stat.st_mtime;
    }

    // remove any old lock files
    time(&t);
    tmP = localtime(&t);
    if (tmP->tm_min == 0) {
      unlinkOldLockFiles();
    }
    
    // "acua expire"
    if (automatic && (tmP->tm_hour == 0) && (tmP->tm_min == 1))
	acua_expire(1, _argv);
    
    // do "acua sync", "acua purge" and "acua expire"
    if (automatic && (tmP->tm_hour == 3)) {
      switch (tmP->tm_min) {
      case 0:
        // sync
        acua_sync(1, _argv);
        break;
      case 1:
        // purge
        if (optPurgeDays >= 14) {
          sprintf(arg, "%d", optPurgeDays);
          _argv[1] = arg;
          acua_purge(2, _argv);
          _argv[1] = NULL;
        }
        break;
//      case 2:
        // expire
//        acua_expire(1, _argv);
//        break;
      }
    }

    // do the update
    update(myHostAddr, t, effect, loggingLevel, timer);
    if (userFile != NULL) {
      userFileClose();
      syslog(LOG_ERR, "Coding error, userbase had to be closed after update()");
    }

    // back to sleep.. yawn.
    time(&t);
    tmP = localtime(&t);
    sleep(60 - tmP->tm_sec);
  }
  return(0);
}

void
update(word myHostAddr, time_t t, int effect, int loggingLevel, int timer)
{
  char			   login[MAX_LOGINCHARS	+ 1],
                           subject[64],
                           nas_exec[256],
                           nas_port[6];
  int			   i,
			   sLeft,
			   curClass,
			   deductMins,
			   deductTime,
			   deductSession,
                           nas_err;
  UserRec		   ur;
  struct radutmp	   rt;
  struct in_addr           __in_addr;
  struct stat              __stat;
  int			   hostIdx=0;


  userFileOpen();

  while	(!userFileRead(&ur)) {
    // make sure the user has a /etc/passwd entry
    if (!loginFromUID(login, ur.uid)) continue;
    // find the index of our host
    hostIdx = findHost(&ur, myHostAddr);
    // warn the user of account expiry.
    WarnExpire(&ur, login);
    // are they online?
    if (ur.nHosts == 0) continue;
    // Ok. We think they're logged in and we're allowed to tinker with 'em.
    // increment tMinutes for this user
    ur.tMinutes++;
    // See if the user is in radutmp, if not then mark them as logged out.
    if (!radiusCheckOnline(&rt, login)) {
      // We were wrong, the user isn't online.

      // remove this host from the users account.
      removeHost(&ur, myHostAddr);
      ur.nHosts = 0;  // dirty hack.
      // update the users acua record..
      if (effect) userFileEdit(&ur);
      // log that we marked them as offline if we actually did (effect)
      if ((loggingLevel >= 1) && (effect))
        syslog(LOG_ERR, "Marked [%s] as logged off (not online).", login);
      continue;
    }
    // this user is online.
    
    if (OVERTIME(ur.flagz) && ((ur.expire - time(NULL)) / 86400) < 1)
    {
    	// turn off the overtime feature as there is not going to be enough
    	// subscription days left to do overtime.
    	ur.flagz &= ~FLG_OVERTIME;
    	
    	// send a email.
    	snprintf(subject, sizeof(subject)-1, "Overtime: Subscription too short.");
        sendMail(&ur, login, "", subject, ETC"/notices/acua_updated.overtime_shortsub");
    }
    
    // Deduct the appropriate amount of time.
	deductMins = 1;
	deductTime = deductSession = 1;
	// get current time class
	curClass = curTimeClass(&ur);
	for (i = 0; i < nTimeClasses; i++) {
          if (inTimeClass(i)) {
    	    if (ur.cLimit[i] <= -3) deductTime = 0;
    	    if (ur.cLimit[i] <= -2) deductSession = 0;
    	    if (ur.cLimit[i] >= 0 && (!TCSMARTTIME(ur.flags)))
              ur.cLeft[i] -= deductMins;
          }
	}
	
	if (timer)
	{
	    if (deductTime) ur.tLeft -= deductMins;
	    if (deductSession) ur.sLeft -= deductMins;
	}
	// figure out how much time they have left
	sLeft = INT_MAX;
	if (ur.tLimit >= 0) sLeft = min(sLeft, ur.tLeft + ur.credit);
	if (ur.sLimit >= 0) sLeft = min(sLeft, ur.sLeft);
	if (curClass >= 0 && ur.cLimit[curClass] >= 0)
          sLeft = min(sLeft, ur.cLeft[curClass]);
	  
	// send a warning message if necessary (about time limit exceeding)
	if (effect && WARNBOOT(ur.flags) && (timer)) {
          if (!OVERTIME(ur.flagz) && (optPPPWarnBoot > 0) && (sLeft == optPPPWarnBoot) && (ur.sLimit < 1440)) {
    	    // set the subject
    	    snprintf(subject, sizeof(subject)-1, "Warning: possible disconnection within %d minutes", optPPPWarnBoot);
    	    // send the e-mail
    	    sendMail(&ur, login, "", subject, ETC"/notices/acua_updated.warn_boot");
    	  }

          if (OVERTIME(ur.flagz) && (optPPPWarnBoot > 0) && (sLeft == optPPPWarnBoot) && (ur.sLimit < 1440)) {
    	    // set the subject
    	    snprintf(subject, sizeof(subject)-1, "Warning: OVERTIME will activate in %d minutes", optPPPWarnBoot);
    	    // send the e-mail
    	    sendMail(&ur, login, "", subject, ETC"/notices/acua_updated.warn_overtime");
          }
	}
	
	/* NEW CODE, OVERTIME option */
	if ((sLeft <= 0) && (OVERTIME(ur.flagz)) && (timer))
	{
	    syslog(LOG_ERR, "Activated OVERTIME feature for [%s]", login);

	    /* make sure they have a day left to knock off (so we dont run into
	     * negative expiry dates) */
	    if (((ur.expire - time(NULL)) / 86400) > 1)
	    {
		syslog(LOG_ERR, "%s", "Looks ok, more than a day left.");
		// deduct a day
		ur.expire -= 86400;
		// Fix time and session remaining.
		ur.tLeft += 60;
		ur.sLeft += 60;
		sLeft += 60;
	    
    		snprintf(subject, sizeof(subject)-1, "Overtime: New expiry date notification.");
    		sendMail(&ur, login, "", subject, ETC"/notices/acua_updated.overtime_newexp");
	    }
	}
	/* NEW CODE, OVERTIME function END */

    if ((sLeft > 0) || ((ur.tLimit >= 1440) && (ur.sLimit >= 1440))) {
      if (effect) userFileEdit(&ur);
      continue;
    }
    
    // update the users record to say when they were last online.
    time(&ur.lastOnline[hostIdx]);

    // is it time for this user to be kicked off?

    // remove us from the users list of hosts that he/she is logged into
    // we do this here instead of below because that way if there's a problem
    // below we wont go getting into a stupid endless loop.
    removeHost(&ur, myHostAddr);
    // update the users account to reflect the changes
    if (effect) userFileEdit(&ur);
    if (loggingLevel >= 1)
      syslog(LOG_INFO, "Marked [%s] as logged out.", login);
    // FIXME: This will fuck up in MultiLink or when maxLogins > 1 and more
    //        than one connection is active, or maybe it wont cause of the
    //        while loop we're actually in at this very moment (not the one
    //        below!).  Study code and fix either this comment or the code.
    //
    // FIXME: from now on the code that handles the NAS's is absolute
    //        shit.  It needs a re-think and a re-write something shocking.
    //
    // Disconnect the user

    /* 07-08-2001 (DD-MM-YYYY) */
    __in_addr.s_addr = rt.nas_address;
    if (snprintf(nas_exec, sizeof(nas_exec), ETC"/nas/%s", inet_ntoa(__in_addr)) == -1) {
      syslog(LOG_ERR, "Could not kick [%s]: %s path truncated.", login, nas_exec);
      continue;
    }

    if (stat(nas_exec, &__stat)) {
      syslog(LOG_ERR, "Could not kick [%s]: %s file not found.", login, nas_exec);
      continue;
    }

    /* Convert nas_port int to a string */
    snprintf(nas_port, sizeof(nas_port), "%d", rt.nas_port);

    /* Kick the user off */
    if ((nas_err = runCommand(nas_exec, login, nas_port, inet_ntoa(__in_addr), NULL))) {
      /* Problem */
      syslog(LOG_ERR, "Could not kick [%s]: %s returned error code %d", login, nas_exec, nas_err);
      continue;
    }

    syslog(LOG_INFO, "Kicked off [%s].", login);
  }
  userFileClose();
}


int
sendMail(UserRec *ur, char *login, char *cc, char *subject, char *sPath)
{
  int   cstat, timer;
  char  *dest, dPath[32];
  pid_t cpid;

  /* Create tempory file */
  snprintf(dPath, sizeof(dPath), "/tmp/acuaXXXXXX");
  if ((mkstemp(dPath)) == -1) {
    /* Problem */
    syslog(LOG_ERR, "Could not create tempory file in /tmp");
    return(-1);
  }

  // make the mail to send the user, from the template file in sPath
  preprocessFile(ur, sPath, dPath, 0);
  // take care of where this mail is going
  if (optMailHost[0] != '\0') {
    dest = (char *) malloc(strlen(optMailHost)+strlen(login)+2);
    if (!dest) errQuit("Not enough memory to hold users email address");
    sprintf(dest, "%s@%s", login, optMailHost);
  } else {
    dest = (char *) malloc(strlen(login)+1);
    if (!dest) errQuit("Not enough memory to hold users email address");
    sprintf(dest, "%s", login);
  }
  
  // fork a new process
  cpid = fork();
  
  if (cpid == 0) {
    // child process - replace stdin with the output from preprocessFile()
    // BTW - changing the uid etc after the freopen is not a bug.
    freopen(dPath, "r", stdin);
    // change to a different user if need be
    if (optMailUser) {
      setuid(optMailUser);   seteuid(optMailUser);
    }
    // change to different group if need be
    if (optMailGroup) {
      setgid(optMailGroup);  setegid(optMailGroup);
    }

    // send the email
    if (cc == NULL) execl(optMailProg, "mail", "-s", subject, dest, NULL); else
	execl(optMailProg, "mail", "-s", subject, "-c", cc, dest, NULL);
    // FIXME: what if execl fails!
  } else if (cpid == -1) {
    syslog(LOG_ERR, "Warning - Could not send mail to [%s], fork failed.", login);
    unlink(dPath);
    // free the malloc'd memory.
    free(dest);
    return(1);
  }
  // this is a safety measure, just incase the mail prog crashes.
  timer = time(NULL);
  for (;;) {
    if (waitpid(cpid, &cstat, WNOHANG) > 0) break;
    if (time(NULL) > timer+10) {
      syslog(LOG_ERR, "Warning - [%s] crashed. Killing dead process", optMailProg);
      kill(cpid, SIGKILL); // go away
      // free the malloc'd memory.
      free(dest);
      return(1);
    }
    usleep(250);
  }
  // free the malloc'd memory.
  free(dest);
  // delete the tempory file we made.
  unlink(dPath);
  return(0);
}

// warn user that they will be expiring soon.
int
warnExpire(time_t expire) {
  int			   i;
  time_t		   t = time(NULL), low,	high;

  for (i = 0; i	< nExpireWarnTimes; i++) {
    low	= expire - expireWarnTime[i] * 24 * 60 * 60;
    high = expire - (expireWarnTime[i] - 1) * 24 * 60 *	60;
    if (t >= low + 120 && t < high) return 1;
  }
  return 0;
}

void
unlinkOldLockFiles()
{
  DIR		 *dir =	opendir(LIB);
  char		  path[256];
  struct dirent	 *_dirent;
  struct stat	  _stat;

  if (dir == NULL)
    perrQuit("opendir(%s)", LIB);

  while	((_dirent = readdir(dir)) != NULL) {
    if (!strncmp(_dirent->d_name, "LOCK", 4)) {
      sprintf(path, "%s/%s", LIB, _dirent->d_name);
      if (lstat(path, &_stat))
	continue;
      if (time(NULL) - _stat.st_atime >= 3600)
	unlink(path);
    }	   
  }
  closedir(dir);
}

void WarnExpire(UserRec *ur, char *login) {
  char subject[128];
  char cc[128];

  cc[0] = '\0';

  // warn users when their accounts/subscriptions are almost expired
  if (ur->expire) {
    if (WARNEXPIRE(ur->flags) && !warnExpire(ur->expire)) {
      ur->flags &= ~FLG_WARNEXPIRE;
      userFileEdit(ur);
    } else
    if (!WARNEXPIRE(ur->flags) && warnExpire(ur->expire) && !NOWARNEXPIRE(ur->flagz)) {
      // set the subject
      if (EXPIRE(ur->flags) == EXPIRE_DELETE)
	strcpy(subject, "Account Expires Soon");
      else
	strcpy(subject, "Subscription Expires Soon");
      // set the CC string
      if (optWarnExpireCC[0] != '\0') {
        sprintf(cc, "-c %s ", optWarnExpireCC);
      }
      // send the e-mail
      sendMail(ur, login, cc, subject, ETC"/notices/acua_updated.warn_expire");
      ur->flags |= FLG_WARNEXPIRE;
      userFileEdit(ur);
    }
  }
}

